﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Problem {
  using System.Collections.Generic;

  using HeuristicLab.Core;
  using HeuristicLab.Encodings.IntegerVectorEncoding;
  using HeuristicLab.Optimization;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Configuration;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Data.Pool;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Extensions;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Generators.CodeGenerator;
  using HeuristicLab.Random;

  public static class IndividualMapper {
    public static PushProgram ToPushProgram(this Individual individual, IReadOnlyPushConfiguration config, ObjectPool<IRandom> randomPool) {
      return individual.IntegerVector().ToPushProgram(config, randomPool);
    }

    public static PushProgram ToPushProgram(this Individual individual, IReadOnlyPushConfiguration config) {
      return individual.IntegerVector().ToPushProgram(config);
    }

    public static PushProgram ToPushProgram(this Individual individual, IReadOnlyPushConfiguration config, IRandom random) {
      return individual.IntegerVector().ToPushProgram(config, random);
    }

    public static PushProgram ToPushProgram(this IntegerVector vector, IReadOnlyPushConfiguration config, ObjectPool<IRandom> randomPool) {
      var random = randomPool.Allocate();
      var seed = vector.GetSeed();

      random.Reset(seed);
      var program = vector.ToPushProgram(config, random);
      randomPool.Free(random);

      return program;
    }

    public static PushProgram ToPushProgram(this IntegerVector vector, IReadOnlyPushConfiguration config) {
      var seed = (uint)vector.GetSeed();
      return vector.ToPushProgram(config, new MersenneTwister(seed));
    }

    public static int GetSeed(this IntegerVector vector) {
      var seed = 17;
      for (var i = 0; i < vector.Length; i++)
        seed = seed * 23 + vector[i];
      return seed;
    }

    public static PushProgram ToPushProgram(this IntegerVector vector, IReadOnlyPushConfiguration config, IRandom random) {
      var currentIndex = 0;
      var close = 0;

      return FromPlush(vector, ref currentIndex, ref close, 0, config, random);
    }

    private static PushProgram FromPlush(IntegerVector vector, ref int currentIndex, ref int close, int depth, IReadOnlyPushConfiguration config, IRandom random) {
      if (currentIndex >= vector.Length)
        return PushProgram.Empty;

      var expressions = new List<Expression>();

      for (; currentIndex < vector.Length; currentIndex++) {
        var expression = CodeGeneratorUtils.CreateExpressionOrErc(
          vector[currentIndex] % config.EnabledExpressions.Count,
          random,
          config.EnabledExpressions,
          config.ErcOptions);

        var expressionType = expression.GetType();

        if (config.ParenthesesCloseBiasLevel > 0) {
          close += random.NextBiased(0, config.MaxParenthesesClose, config.ParenthesesCloseBiasLevel);

          // check if expression requires additional blocks
          if (ExpressionTable.TypeToAttributeTable.ContainsKey(expressionType)) {
            var attr = ExpressionTable.TypeToAttributeTable[expressionType];

            for (var blockIdx = 0u; blockIdx < attr.ExecIn && currentIndex < vector.Length; blockIdx++) {
              if (close != 0) {
                close--;
                expressions.Add(PushProgram.Empty);
              } else {
                currentIndex++;
                var subProgram = FromPlush(vector, ref currentIndex, ref close, depth + 1, config, random);
                var subExpression = subProgram.Count == 1 ? subProgram.Expressions[0] : subProgram;
                expressions.Add(subExpression);
              }
            }
          }
        }

        expressions.Add(expression);

        if (close > 0 && depth > 0) {
          close--;
          break;
        }

        if (depth == 0) {
          close = 0;
        }
      }

      return new PushProgram(expressions);
    }
  }
}
